@@ -20,9 +20,6 @@
 #ifndef GDB_DWARF2_READ_H
 #define GDB_DWARF2_READ_H
 
-#if CXX_STD_THREAD
-#include <mutex>
-#endif
 #include <queue>
 #include "dwarf2/abbrev.h"
 #include "dwarf2/unit-head.h"
@@ -32,6 +29,7 @@
 #include "dwarf2/section.h"
 #include "dwarf2/cu.h"
 #include "dwarf2/dwz.h"
+#include "gdbsupport/cxx-thread.h"
 #include "gdbsupport/gdb_obstack.h"
 #include "gdbsupport/function-view.h"
 #include "gdbsupport/packed.h"
@@ -54,6 +52,7 @@ struct dwarf2_cu;
 struct dwarf2_debug_sections;
 struct dwarf2_per_bfd;
 struct dwarf2_per_cu;
+struct file_entry;
 struct mapped_index;
 struct mapped_debug_names;
 struct signatured_type;
@@ -117,48 +116,46 @@ struct dwarf2_per_cu
   /* LENGTH is the length of the unit.  If the value is 0, it means it is not
      known, and may be set later using the set_length method.  */
   dwarf2_per_cu (dwarf2_per_bfd *per_bfd, dwarf2_section_info *section,
-		 sect_offset sect_off, unsigned int length, bool is_dwz)
-    : sect_off (sect_off),
+		 sect_offset sect_off, unsigned int length, bool is_dwz,
+		 bool is_debug_types = false)
+    : m_sect_off (sect_off),
       m_length (length),
-      is_debug_types (false),
-      is_dwz (is_dwz),
+      m_is_debug_types (is_debug_types),
+      m_is_dwz (is_dwz),
       reading_dwo_directly (false),
       tu_read (false),
       lto_artificial (false),
       queued (false),
-      mark (false),
+      m_header_read_in (false),
       files_read (false),
       scanned (false),
-      section (section),
-      per_bfd (per_bfd)
+      m_section (section),
+      m_per_bfd (per_bfd)
   {
     gdb_assert (per_bfd != nullptr);
     gdb_assert (section != nullptr);
   }
 
-  /* The start offset of this unit.
-
-     If the unit is located in a DWO file, and has a corresponding skeleton
-     unit in the main file, this is the offset of the skeleton unit in its
-     containing section in the main file.
-
-     If the unit is located in a DWO file, but has no corresponding skeleton
-     unit in the main file, this is the offset of the unit in its containing
-     section in the DWO file.  */
-  sect_offset sect_off;
-
 private:
+  /* The start offset and length of this compilation unit.
+     NOTE: Unlike comp_unit_head.length, this length includes
+     initial_length_size.
+     If the DIE refers to a DWO file, this is always of the original die,
+     not the DWO file.  */
+  sect_offset m_sect_off;
+
   unsigned int m_length = 0;
 
-public:
+private:
   /* Non-zero if this CU is from .debug_types.
      Struct dwarf2_per_cu is contained in struct signatured_type iff
      this is non-zero.  */
-  unsigned int is_debug_types : 1;
+  unsigned int m_is_debug_types : 1;
 
   /* Non-zero if this CU is from the .dwz file.  */
-  unsigned int is_dwz : 1;
+  unsigned int m_is_dwz : 1;
 
+public:
   /* Non-zero if reading a TU directly from a DWO file, bypassing the stub.
      This flag is only valid if is_debug_types is true.
      We can't read a CU directly from a DWO file: There are required
@@ -192,10 +189,11 @@ public:
      any of the current compilation units are processed.  */
   packed<bool, 1> queued;
 
-  /* A temporary mark bit used when iterating over all CUs in
-     expand_symtabs_matching.  */
-  packed<unsigned int, 1> mark;
+private:
+  /* True if HEADER has been read in.  */
+  mutable packed<bool, 1> m_header_read_in;
 
+public:
   /* True if we've tried to read the file table.  There will be no
      point in trying to read it again next time.  */
   packed<bool, 1> files_read;
@@ -221,31 +219,25 @@ public:
      not.  */
   std::atomic<bool> scanned;
 
-private:
-  /* Sizes for an address, an offset, and a section offset.  These fields are
-     set by cutu_reader when the unit is read.  */
-  std::atomic<std::uint8_t> m_addr_size = 0;
-  std::atomic<std::uint8_t> m_offset_size = 0;
-  std::atomic<std::uint8_t> m_ref_addr_size = 0;
-
-public:
   /* Our index in the unshared "symtabs" vector.  */
   unsigned index = 0;
 
-  /* The section this unit lives in.
-
-     If the unit is located in a DWO file, and has a corresponding skeleton unit
-     in the main file, this is the section of the skeleton unit in the main
-     file.
-
-     If the unit is located in a DWO file, but has no corresponding skeleton
-     unit in the main file, this is the section of the unit in the DWO
-     file.  */
-  struct dwarf2_section_info *section = nullptr;
+private:
+  /* The section this CU/TU lives in.
+     If the DIE refers to a DWO file, this is always the original die,
+     not the DWO file.  */
+  dwarf2_section_info *m_section;
 
   /* Backlink to the owner of this.  */
-  dwarf2_per_bfd *per_bfd;
+  dwarf2_per_bfd *m_per_bfd;
 
+  /* DWARF header of this unit.  Note that dwarf2_cu reads its own version of
+     the header, which may differ from this one, since it may pass
+     rch_kind::TYPE to read_unit_head, whereas for dwarf2_per_cu we always pass
+     ruh_kind::COMPILE.  */
+  mutable unit_head m_header;
+
+public:
   /* The file and directory for this CU.  This is cached so that we
      don't need to re-examine the DWO in some situations.  This may be
      nullptr, depending on the CU; for example a partial unit won't
@@ -274,53 +266,35 @@ public:
      http://sourceware.org/bugzilla/show_bug.cgi?id=15021.  */
   std::vector<dwarf2_per_cu *> imported_symtabs;
 
+  bool is_debug_types () const
+  { return m_is_debug_types; }
+
+  dwarf2_per_bfd *per_bfd () const
+  { return m_per_bfd; }
+
+  dwarf2_section_info *section () const
+  { return m_section; }
+
+  sect_offset sect_off () const
+  { return m_sect_off; }
+
+  bool is_dwz () const
+  { return m_is_dwz; }
+
+  /* Get the header of this per_cu, reading it if necessary.  */
+  const unit_head *get_header () const;
+
   /* Return the address size given in the compilation unit header for
      this CU.  */
-  std::uint8_t addr_size () const
-  {
-    gdb_assert (m_addr_size != 0);
-    return m_addr_size;
-  }
-
-  /* Set the address size given in the compilation unit header for
-     this CU.  */
-  void set_addr_size (std::uint8_t addr_size)
-  {
-    std::uint8_t old = m_addr_size.exchange (addr_size);
-    gdb_assert (old == 0 || old == addr_size);
-  }
+  int addr_size () const;
 
   /* Return the offset size given in the compilation unit header for
      this CU.  */
-  std::uint8_t offset_size () const
-  {
-    gdb_assert (m_offset_size != 0);
-    return m_offset_size;
-  }
-
-  /* Set the offset size given in the compilation unit header for
-     this CU.  */
-  void set_offset_size (std::uint8_t offset_size)
-  {
-    std::uint8_t old = m_offset_size.exchange (offset_size);
-    gdb_assert (old == 0 || old == offset_size);
-  }
+  int offset_size () const;
 
   /* Return the DW_FORM_ref_addr size given in the compilation unit
      header for this CU.  */
-  std::uint8_t ref_addr_size () const
-  {
-    gdb_assert (m_ref_addr_size != 0);
-    return m_ref_addr_size;
-  }
-
-  /* Set the DW_FORM_ref_addr size given in the compilation unit
-     header for this CU.  */
-  void set_ref_addr_size (std::uint8_t ref_addr_size)
-  {
-    std::uint8_t old = m_ref_addr_size.exchange (ref_addr_size);
-    gdb_assert (old == 0 || old == ref_addr_size);
-  }
+  int ref_addr_size () const;
 
   /* Return length of this CU.  */
   unsigned int length () const
@@ -415,11 +389,9 @@ struct signatured_type : public dwarf2_per_cu
   signatured_type (dwarf2_per_bfd *per_bfd, dwarf2_section_info *section,
 		   sect_offset sect_off, unsigned int length, bool is_dwz,
 		   ULONGEST signature)
-    : dwarf2_per_cu (per_bfd, section, sect_off, length, is_dwz),
+    : dwarf2_per_cu (per_bfd, section, sect_off, length, is_dwz, true),
       signature (signature)
-  {
-    this->is_debug_types = true;
-  }
+  {}
 
   /* The type's signature.  */
   ULONGEST signature;
@@ -639,8 +611,10 @@ public:
 
   std::vector<dwarf2_section_info> types;
 
-  /* Table of all the compilation units.  This is used to locate
-     the target compilation unit of a particular reference.  */
+  /* Table of all compilation and type units.
+
+     The order of this vector matters, because it is used to locate targets of
+     DW_FORM_ref_addr attributes (reference by section offset).  */
   std::vector<dwarf2_per_cu_up> all_units;
 
   /* Number of compilation and type units in the ALL_UNITS vector.  */
@@ -657,10 +631,8 @@ public:
   /* Set of dwo_file objects.  */
   dwo_file_up_set dwo_files;
 
-#if CXX_STD_THREAD
   /* Mutex to synchronize access to DWO_FILES.  */
-  std::mutex dwo_files_lock;
-#endif
+  gdb::mutex dwo_files_lock;
 
   /* The DWP file if there is one, or NULL.  */
   dwp_file_up dwp_file;
@@ -1257,25 +1229,44 @@ struct dwarf2_base_index_functions : public quick_symbol_functions
 			     bool need_fullname) override;
 };
 
-/* If FILE_MATCHER is NULL or if PER_CU has
-   dwarf2_per_cu_quick_data::MARK set (see
-   dw_expand_symtabs_matching_file_matcher), expand the CU and call
-   EXPANSION_NOTIFY on it.  */
+/* This is used to track whether a CU has already been visited during
+   symbol expansion.  It is an auto-resizing bool vector.  */
+class auto_bool_vector
+{
+public:
 
-extern bool dw2_expand_symtabs_matching_one
+  auto_bool_vector () = default;
+
+  /* Return true if element I is set.  */
+  bool is_set (size_t i) const
+  {
+    if (i < m_vec.size ())
+      return m_vec[i];
+    return false;
+  }
+
+  /* Set a value in this vector, growing it automatically.  */
+  void set (size_t i, bool value)
+  {
+    if (m_vec.size () < i + 1)
+      m_vec.resize (i + 1);
+    m_vec[i] = value;
+  }
+
+private:
+  std::vector<bool> m_vec;
+};
+
+/* If FILE_MATCHER is NULL and if CUS_TO_SKIP does not include the
+   CU's index, expand the CU and call LISTENER on it.  */
+
+extern bool dw2_search_one
   (dwarf2_per_cu *per_cu,
    dwarf2_per_objfile *per_objfile,
-   expand_symtabs_file_matcher file_matcher,
-   expand_symtabs_expansion_listener expansion_notify,
-   expand_symtabs_lang_matcher lang_matcher);
-
-/* If FILE_MATCHER is non-NULL, set all the
-   dwarf2_per_cu_quick_data::MARK of the current DWARF2_PER_OBJFILE
-   that match FILE_MATCHER.  */
-
-extern void dw_expand_symtabs_matching_file_matcher
-  (dwarf2_per_objfile *per_objfile,
-   expand_symtabs_file_matcher file_matcher);
+   auto_bool_vector &cus_to_skip,
+   search_symtabs_file_matcher file_matcher,
+   search_symtabs_expansion_listener listener,
+   search_symtabs_lang_matcher lang_matcher);
 
 /* Return pointer to string at .debug_str offset STR_OFFSET.  */
 
@@ -1386,4 +1377,40 @@ extern file_and_directory &find_file_and_directory (die_info *die,
 extern const dwarf2_section_info &get_section_for_ref
   (const attribute &attr, dwarf2_cu *cu);
 
+/* A convenience function to find the proper .debug_line section for a CU.  */
+
+extern struct dwarf2_section_info *get_debug_line_section
+  (struct dwarf2_cu *cu);
+
+/* Start a subfile for DWARF.  FILENAME is the name of the file and
+   DIRNAME the name of the source directory which contains FILENAME
+   or NULL if not known.
+   This routine tries to keep line numbers from identical absolute and
+   relative file names in a common subfile.
+
+   Using the `list' example from the GDB testsuite, which resides in
+   /srcdir and compiling it with Irix6.2 cc in /compdir using a filename
+   of /srcdir/list0.c yields the following debugging information for list0.c:
+
+   DW_AT_name:          /srcdir/list0.c
+   DW_AT_comp_dir:      /compdir
+   files.files[0].name: list0.h
+   files.files[0].dir:  /srcdir
+   files.files[1].name: list0.c
+   files.files[1].dir:  /srcdir
+
+   The line number information for list0.c has to end up in a single
+   subfile, so that `break /srcdir/list0.c:1' works as expected.
+   start_subfile will ensure that this happens provided that we pass the
+   concatenation of files.files[1].dir and files.files[1].name as the
+   subfile's name.  */
+extern void dwarf2_start_subfile (dwarf2_cu *cu, const file_entry &fe,
+				  const line_header &lh);
+
+/* A helper function that decides if a given symbol is an Ada Pragma
+   Import or Pragma Export.  */
+
+extern bool is_ada_import_or_export (dwarf2_cu *cu, const char *name,
+				     const char *linkagename);
+
 #endif /* GDB_DWARF2_READ_H */
